@@ -1,8 +1,17 @@
-from typing import List, Literal, Union
+from typing import List, Literal, Union, overload
 
 import numpy
 
+from pyquda_comm.field import (
+    LatticeInfo,
+    LatticeGauge,
+    LatticeMom,
+    LatticeFermion,
+    LatticeStaggeredFermion,
+    LatticeReal,
+)
 from ..pyquda import (
+    QudaEigParam,
     QudaGaugeSmearParam,
     QudaGaugeObservableParam,
     MatQuda,
@@ -23,22 +32,12 @@ from ..pyquda import (
     computeGaugeFixingOVRQuda,
     computeGaugeFixingFFTQuda,
 )
-from ..field import (
-    LatticeInfo,
-    LatticeGauge,
-    LatticeMom,
-    LatticeFermion,
-    LatticeStaggeredFermion,
-    LatticeReal,
-)
 from ..enum_quda import (
     QudaBoolean,
     QudaDslashType,
     QudaGaugeSmearType,
     QudaLinkType,
     QudaMassNormalization,
-    QudaPrecision,
-    QudaReconstructType,
     QudaSolveType,
 )
 
@@ -51,24 +50,14 @@ class GaugeDirac(Dirac):
         super().__init__(LatticeInfo(latt_info.global_size))  # Keep periodic t boundary and isotropic
         self.newQudaGaugeParam()
         self.newQudaInvertParam()
+        self.newQudaEigParam()
         self.newQudaGaugeSmearParam()
         self.newQudaGaugeObservableParam()
-        self.setPrecision(
-            cuda=QudaPrecision.QUDA_DOUBLE_PRECISION,
-            sloppy=QudaPrecision.QUDA_DOUBLE_PRECISION,
-            precondition=QudaPrecision.QUDA_DOUBLE_PRECISION,
-            eigensolver=QudaPrecision.QUDA_DOUBLE_PRECISION,
-        )
-        # Use QUDA_RECONSTRUCT_NO to ensure slight deviations from SU(3) can be preserved
-        self.setReconstruct(
-            cuda=QudaReconstructType.QUDA_RECONSTRUCT_NO,
-            sloppy=QudaReconstructType.QUDA_RECONSTRUCT_NO,
-            precondition=QudaReconstructType.QUDA_RECONSTRUCT_NO,
-            eigensolver=QudaReconstructType.QUDA_RECONSTRUCT_NO,
-        )
+        self.setPrecision()
+        self.setReconstruct()
 
     def newQudaGaugeParam(self):
-        gauge_param = general.newQudaGaugeParam(self.latt_info, 1.0, 0.0)
+        gauge_param = general.newQudaGaugeParam(self.latt_info)
         self.gauge_param = gauge_param
 
     def newQudaInvertParam(self):
@@ -77,6 +66,13 @@ class GaugeDirac(Dirac):
         invert_param.mass_normalization = QudaMassNormalization.QUDA_KAPPA_NORMALIZATION
         self.invert_param = invert_param
 
+    def newQudaEigParam(self):
+        eig_param = QudaEigParam()
+        eig_param.vec_infile = b""
+        eig_param.vec_outfile = b""
+        eig_param.invert_param = self.invert_param
+        self.eig_param = eig_param
+
     def newQudaGaugeSmearParam(self):
         smear_param = QudaGaugeSmearParam()
         self.smear_param = smear_param
@@ -133,6 +129,11 @@ class GaugeDirac(Dirac):
         MatQuda(b.data_ptr, x.data_ptr, self.invert_param)
         return b
 
+    @overload
+    def wuppertalSmear(self, x: LatticeFermion, n_steps: int, alpha: float) -> LatticeFermion: ...
+    @overload
+    def wuppertalSmear(self, x: LatticeStaggeredFermion, n_steps: int, alpha: float) -> LatticeStaggeredFermion: ...
+
     def wuppertalSmear(self, x: Union[LatticeFermion, LatticeStaggeredFermion], n_steps: int, alpha: float):
         if isinstance(x, LatticeStaggeredFermion):
             b = LatticeStaggeredFermion(x.latt_info)
@@ -175,12 +176,10 @@ class GaugeDirac(Dirac):
                 raise ValueError(f"path should be list of int from 0 to 7, but get {path}")
         return input_path_buf, len(path)
 
-    def path(
-        self,
-        gauge: LatticeGauge,
-        paths: List[List[int]],
-    ):
-        gauge_path = LatticeGauge(gauge.latt_info)
+    def path(self, gauge: LatticeGauge, paths: List[List[int]]):
+        from ..field import LatticeGauge as LatticeGauge_
+
+        gauge_path = LatticeGauge_(gauge.latt_info)
         num_paths = 1
         input_path_buf_x, path_length = GaugeDirac._getPath(paths[0])
         input_path_buf = numpy.zeros((4, 1, path_length + 1), "<i4")
@@ -236,13 +235,10 @@ class GaugeDirac(Dirac):
                 raise ValueError(f"path {loops[i]} is not a loop")
         return input_path_buf, path_length, num_paths, max_length
 
-    def loop(
-        self,
-        gauge: LatticeGauge,
-        loops: List[List[List[int]]],
-        coeff: List[float],
-    ):
-        gauge_loop = LatticeGauge(gauge.latt_info)
+    def loop(self, gauge: LatticeGauge, loops: List[List[List[int]]], coeff: List[float]):
+        from ..field import LatticeGauge as LatticeGauge_
+
+        gauge_loop = LatticeGauge_(gauge.latt_info)
         input_path_buf_x, path_length, num_paths, max_length = GaugeDirac._getLoops(loops[0])
         input_path_buf = numpy.zeros((4, num_paths, max_length + 1), "<i4")
         input_path_buf[0, :, 0] = 7
@@ -287,27 +283,54 @@ class GaugeDirac(Dirac):
         )
         return traces
 
-    def apeSmear(self, n_steps: int, alpha: float, dir_ignore: int):
+    def apeSmear(
+        self,
+        n_steps: int,
+        alpha: float,
+        dir_ignore: int,
+        compute_plaquette: bool = False,
+        compute_qcharge: bool = False,
+    ):
         self.smear_param.smear_type = QudaGaugeSmearType.QUDA_GAUGE_SMEAR_APE
         self.smear_param.n_steps = n_steps
         self.smear_param.alpha = alpha
         self.smear_param.meas_interval = n_steps + 1
         self.smear_param.dir_ignore = dir_ignore
-        self.obs_param.compute_qcharge = QudaBoolean.QUDA_BOOLEAN_TRUE
+        self.obs_param.compute_plaquette = QudaBoolean(compute_plaquette)
+        self.obs_param.compute_qcharge = QudaBoolean(compute_qcharge)
         performGaugeSmearQuda(self.smear_param, self.obs_param)
+        self.obs_param.compute_plaquette = QudaBoolean.QUDA_BOOLEAN_FALSE
         self.obs_param.compute_qcharge = QudaBoolean.QUDA_BOOLEAN_FALSE
 
-    def stoutSmear(self, n_steps: int, rho: float, dir_ignore: int):
+    def stoutSmear(
+        self,
+        n_steps: int,
+        rho: float,
+        dir_ignore: int,
+        compute_plaquette: bool = False,
+        compute_qcharge: bool = False,
+    ):
         self.smear_param.smear_type = QudaGaugeSmearType.QUDA_GAUGE_SMEAR_STOUT
         self.smear_param.n_steps = n_steps
         self.smear_param.rho = rho
         self.smear_param.meas_interval = n_steps + 1
         self.smear_param.dir_ignore = dir_ignore
-        self.obs_param.compute_qcharge = QudaBoolean.QUDA_BOOLEAN_TRUE
+        self.obs_param.compute_plaquette = QudaBoolean(compute_plaquette)
+        self.obs_param.compute_qcharge = QudaBoolean(compute_qcharge)
         performGaugeSmearQuda(self.smear_param, self.obs_param)
+        self.obs_param.compute_plaquette = QudaBoolean.QUDA_BOOLEAN_FALSE
         self.obs_param.compute_qcharge = QudaBoolean.QUDA_BOOLEAN_FALSE
 
-    def hypSmear(self, n_steps: int, alpha1: float, alpha2: float, alpha3: float, dir_ignore: int):
+    def hypSmear(
+        self,
+        n_steps: int,
+        alpha1: float,
+        alpha2: float,
+        alpha3: float,
+        dir_ignore: int,
+        compute_plaquette: bool = False,
+        compute_qcharge: bool = False,
+    ):
         self.smear_param.smear_type = QudaGaugeSmearType.QUDA_GAUGE_SMEAR_HYP
         self.smear_param.n_steps = n_steps
         self.smear_param.alpha1 = alpha1
@@ -315,32 +338,50 @@ class GaugeDirac(Dirac):
         self.smear_param.alpha3 = alpha3
         self.smear_param.meas_interval = n_steps + 1
         self.smear_param.dir_ignore = dir_ignore
-        self.obs_param.compute_qcharge = QudaBoolean.QUDA_BOOLEAN_TRUE
+        self.obs_param.compute_plaquette = QudaBoolean(compute_plaquette)
+        self.obs_param.compute_qcharge = QudaBoolean(compute_qcharge)
         performGaugeSmearQuda(self.smear_param, self.obs_param)
+        self.obs_param.compute_plaquette = QudaBoolean.QUDA_BOOLEAN_FALSE
         self.obs_param.compute_qcharge = QudaBoolean.QUDA_BOOLEAN_FALSE
 
-    def wilsonFlow(self, n_steps: int, epsilon: float, t0: float, restart: bool):
+    def wilsonFlow(
+        self,
+        n_steps: int,
+        epsilon: float,
+        t0: float,
+        restart: bool,
+        compute_plaquette: bool = False,
+        compute_qcharge: bool = True,
+    ):
         self.smear_param.smear_type = QudaGaugeSmearType.QUDA_GAUGE_SMEAR_WILSON_FLOW
         self.smear_param.n_steps = n_steps
         self.smear_param.epsilon = epsilon
         self.smear_param.t0 = t0
         self.smear_param.restart = QudaBoolean(restart)
         self.smear_param.meas_interval = n_steps + 1
-        self.obs_param.compute_plaquette = QudaBoolean.QUDA_BOOLEAN_TRUE
-        self.obs_param.compute_qcharge = QudaBoolean.QUDA_BOOLEAN_TRUE
+        self.obs_param.compute_plaquette = QudaBoolean(compute_plaquette)
+        self.obs_param.compute_qcharge = QudaBoolean(compute_qcharge)
         performWFlowQuda(self.smear_param, self.obs_param)
         self.obs_param.compute_plaquette = QudaBoolean.QUDA_BOOLEAN_FALSE
         self.obs_param.compute_qcharge = QudaBoolean.QUDA_BOOLEAN_FALSE
 
-    def symanzikFlow(self, n_steps: int, epsilon: float, t0: float, restart: bool):
+    def symanzikFlow(
+        self,
+        n_steps: int,
+        epsilon: float,
+        t0: float,
+        restart: bool,
+        compute_plaquette: bool = False,
+        compute_qcharge: bool = True,
+    ):
         self.smear_param.smear_type = QudaGaugeSmearType.QUDA_GAUGE_SMEAR_SYMANZIK_FLOW
         self.smear_param.n_steps = n_steps
         self.smear_param.epsilon = epsilon
         self.smear_param.t0 = t0
         self.smear_param.restart = QudaBoolean(restart)
         self.smear_param.meas_interval = n_steps + 1
-        self.obs_param.compute_plaquette = QudaBoolean.QUDA_BOOLEAN_TRUE
-        self.obs_param.compute_qcharge = QudaBoolean.QUDA_BOOLEAN_TRUE
+        self.obs_param.compute_plaquette = QudaBoolean(compute_plaquette)
+        self.obs_param.compute_qcharge = QudaBoolean(compute_qcharge)
         performWFlowQuda(self.smear_param, self.obs_param)
         self.obs_param.compute_plaquette = QudaBoolean.QUDA_BOOLEAN_FALSE
         self.obs_param.compute_qcharge = QudaBoolean.QUDA_BOOLEAN_FALSE
@@ -371,7 +412,7 @@ class GaugeDirac(Dirac):
 
     def qchargeDensity(self):
         qcharge_density = LatticeReal(self.latt_info)
-        self.obs_param.qcharge_density = qcharge_density.data_ptr
+        self.obs_param.qcharge_density = qcharge_density.data_void_ptr
         self.obs_param.compute_qcharge_density = QudaBoolean.QUDA_BOOLEAN_TRUE
         gaugeObservablesQuda(self.obs_param)
         self.obs_param.compute_qcharge_density = QudaBoolean.QUDA_BOOLEAN_TRUE
