@@ -1,9 +1,16 @@
-from typing import List, NamedTuple
+from typing import List, Literal, NamedTuple, Optional
 
 import numpy
 from numpy.typing import NDArray
 
-from ..pointer import Pointer, Pointers
+from pyquda_comm import getLogger, getArrayBackendTarget
+from ..field import (
+    LatticeInfo,
+    LatticeGauge,
+    LatticeClover,
+    LatticeFermion,
+    LatticeStaggeredFermion,
+)
 from ..pyquda import (
     QudaGaugeParam,
     QudaInvertParam,
@@ -18,13 +25,6 @@ from ..pyquda import (
     computeKSLinkQuda,
     staggeredPhaseQuda,
 )
-from ..field import (
-    LatticeInfo,
-    LatticeGauge,
-    LatticeClover,
-    LatticeFermion,
-    LatticeStaggeredFermion,
-)
 from ..enum_quda import (
     QUDA_MAX_DIM,
     QUDA_MAX_MG_LEVEL,
@@ -58,9 +58,10 @@ from ..enum_quda import (
     QudaBoolean,
     QudaStaggeredPhase,
 )
+from ..quda_define import mmaAvailable
 
-nullptr = Pointer("void")
-nullptrs = Pointers("void", 0)
+nullptr = numpy.empty((0), "<c16")
+nullptrs = numpy.empty((0, 0), "<c16")
 
 
 class Precision(NamedTuple):
@@ -78,83 +79,102 @@ class Reconstruct(NamedTuple):
     eigensolver: QudaReconstructType
 
 
-_precision = Precision(
-    QudaPrecision.QUDA_DOUBLE_PRECISION,
-    QudaPrecision.QUDA_DOUBLE_PRECISION,
-    QudaPrecision.QUDA_HALF_PRECISION,
-    QudaPrecision.QUDA_HALF_PRECISION,
-    QudaPrecision.QUDA_SINGLE_PRECISION,
-)
-_reconstruct = Reconstruct(
-    QudaReconstructType.QUDA_RECONSTRUCT_12,
-    QudaReconstructType.QUDA_RECONSTRUCT_12,
-    QudaReconstructType.QUDA_RECONSTRUCT_12,
-    QudaReconstructType.QUDA_RECONSTRUCT_12,
-)
+_precision = {
+    "none": Precision(
+        QudaPrecision.QUDA_DOUBLE_PRECISION,
+        QudaPrecision.QUDA_DOUBLE_PRECISION,
+        QudaPrecision.QUDA_DOUBLE_PRECISION,
+        QudaPrecision.QUDA_DOUBLE_PRECISION,
+        QudaPrecision.QUDA_DOUBLE_PRECISION,
+    ),
+    "invert": Precision(
+        QudaPrecision.QUDA_DOUBLE_PRECISION,
+        QudaPrecision.QUDA_DOUBLE_PRECISION,
+        QudaPrecision.QUDA_HALF_PRECISION,
+        QudaPrecision.QUDA_HALF_PRECISION,
+        QudaPrecision.QUDA_DOUBLE_PRECISION,
+    ),
+    "multigrid": Precision(
+        QudaPrecision.QUDA_DOUBLE_PRECISION,
+        QudaPrecision.QUDA_DOUBLE_PRECISION,
+        QudaPrecision.QUDA_SINGLE_PRECISION,
+        QudaPrecision.QUDA_HALF_PRECISION,
+        QudaPrecision.QUDA_DOUBLE_PRECISION,
+    ),
+}
+_reconstruct = {
+    "none": Reconstruct(
+        QudaReconstructType.QUDA_RECONSTRUCT_NO,
+        QudaReconstructType.QUDA_RECONSTRUCT_NO,
+        QudaReconstructType.QUDA_RECONSTRUCT_NO,
+        QudaReconstructType.QUDA_RECONSTRUCT_NO,
+    ),
+    "wilson": Reconstruct(
+        QudaReconstructType.QUDA_RECONSTRUCT_12,
+        QudaReconstructType.QUDA_RECONSTRUCT_8,
+        QudaReconstructType.QUDA_RECONSTRUCT_8,
+        QudaReconstructType.QUDA_RECONSTRUCT_12,
+    ),
+    "staggered": Reconstruct(
+        QudaReconstructType.QUDA_RECONSTRUCT_13,
+        QudaReconstructType.QUDA_RECONSTRUCT_9,
+        QudaReconstructType.QUDA_RECONSTRUCT_9,
+        QudaReconstructType.QUDA_RECONSTRUCT_13,
+    ),
+}
 
 
-def getGlobalPrecision():
-    return _precision
-
-
-def getGlobalReconstruct():
-    return _reconstruct
+def getGlobalPrecision(key: Literal["none", "invert", "multigrid"]):
+    return _precision[key]
 
 
 def setGlobalPrecision(
+    key: Literal["none", "invert", "multigrid"],
     *,
-    cuda: QudaPrecision = None,
-    sloppy: QudaPrecision = None,
-    precondition: QudaPrecision = None,
-    eigensolver: QudaPrecision = None,
+    cuda: Optional[QudaPrecision] = None,
+    sloppy: Optional[QudaPrecision] = None,
+    precondition: Optional[QudaPrecision] = None,
+    eigensolver: Optional[QudaPrecision] = None,
 ):
     global _precision
-    _precision = Precision(
-        _precision.cpu,
-        cuda if cuda is not None else _precision.cuda,
-        sloppy if sloppy is not None else _precision.sloppy,
-        precondition if precondition is not None else _precision.precondition,
-        eigensolver if eigensolver is not None else _precision.eigensolver,
+    precision = _precision[key]
+    _precision[key] = Precision(
+        precision.cpu,
+        cuda if cuda is not None else precision.cuda,
+        sloppy if sloppy is not None else precision.sloppy,
+        precondition if precondition is not None else precision.precondition,
+        eigensolver if eigensolver is not None else precision.eigensolver,
     )
 
 
+def getGlobalReconstruct(key: Literal["none", "wilson", "staggered"]):
+    return _reconstruct[key]
+
+
 def setGlobalReconstruct(
+    key: Literal["none", "wilson", "staggered"],
     *,
-    cuda: QudaReconstructType = None,
-    sloppy: QudaReconstructType = None,
-    precondition: QudaReconstructType = None,
-    eigensolver: QudaReconstructType = None,
+    cuda: Optional[QudaReconstructType] = None,
+    sloppy: Optional[QudaReconstructType] = None,
+    precondition: Optional[QudaReconstructType] = None,
+    eigensolver: Optional[QudaReconstructType] = None,
 ):
     global _reconstruct
-    _reconstruct = Reconstruct(
-        cuda if cuda is not None else _reconstruct.cuda,
-        sloppy if sloppy is not None else _reconstruct.sloppy,
-        precondition if precondition is not None else _reconstruct.precondition,
-        eigensolver if eigensolver is not None else _reconstruct.eigensolver,
+    reconstruct = _reconstruct[key]
+    _reconstruct[key] = Reconstruct(
+        cuda if cuda is not None else reconstruct.cuda,
+        sloppy if sloppy is not None else reconstruct.sloppy,
+        precondition if precondition is not None else reconstruct.precondition,
+        eigensolver if eigensolver is not None else reconstruct.eigensolver,
     )
 
 
-def _fieldLocation():
-    from .. import getCUDABackend
-
-    if getCUDABackend() == "numpy":
-        return QudaFieldLocation.QUDA_CPU_FIELD_LOCATION
-    else:
-        return QudaFieldLocation.QUDA_CUDA_FIELD_LOCATION
-
-
-def _useMMA():
-    from .. import isHIP, getCUDAComputeCapability
-
-    return QudaBoolean(not isHIP() and getCUDAComputeCapability().major >= 7)
-
-
 def setPrecisionParam(
     precision: Precision,
-    gauge_param: QudaGaugeParam = None,
-    invert_param: QudaInvertParam = None,
-    mg_param: QudaMultigridParam = None,
-    mg_inv_param: QudaInvertParam = None,
+    gauge_param: Optional[QudaGaugeParam] = None,
+    invert_param: Optional[QudaInvertParam] = None,
+    mg_param: Optional[QudaMultigridParam] = None,
+    mg_inv_param: Optional[QudaInvertParam] = None,
 ):
     if gauge_param is not None:
         gauge_param.cpu_prec = precision.cpu
@@ -185,17 +205,19 @@ def setPrecisionParam(
         mg_inv_param.cuda_prec = precision.cuda
         mg_inv_param.cuda_prec_sloppy = precision.sloppy
         mg_inv_param.cuda_prec_precondition = precision.precondition
+        mg_inv_param.cuda_prec_eigensolver = precision.eigensolver
 
         mg_inv_param.clover_cpu_prec = precision.cpu
         mg_inv_param.clover_cuda_prec = precision.cuda
         mg_inv_param.clover_cuda_prec_sloppy = precision.sloppy
         mg_inv_param.clover_cuda_prec_precondition = precision.precondition
+        mg_inv_param.clover_cuda_prec_eigensolver = precision.eigensolver
 
     if mg_param is not None:
         mg_param.precision_null = [precision.precondition] * QUDA_MAX_MG_LEVEL
 
 
-def setReconstructParam(reconstruct: Reconstruct, gauge_param: QudaGaugeParam = None):
+def setReconstructParam(reconstruct: Reconstruct, gauge_param: Optional[QudaGaugeParam] = None):
     if gauge_param is not None:
         gauge_param.reconstruct = reconstruct.cuda
         gauge_param.reconstruct_sloppy = reconstruct.sloppy
@@ -204,21 +226,24 @@ def setReconstructParam(reconstruct: Reconstruct, gauge_param: QudaGaugeParam =
         gauge_param.reconstruct_eigensolver = reconstruct.eigensolver
 
 
-def newQudaGaugeParam(
-    lattice: LatticeInfo,
-    tadpole_coeff: float,
-    naik_epsilon: float,
-):
+def fieldLocation() -> QudaFieldLocation:
+    if getArrayBackendTarget() == "cpu":
+        return QudaFieldLocation.QUDA_CPU_FIELD_LOCATION
+    else:
+        return QudaFieldLocation.QUDA_CUDA_FIELD_LOCATION
+
+
+def newQudaGaugeParam(lattice: LatticeInfo):
     gauge_param = QudaGaugeParam()
 
     gauge_param.X = lattice.size
 
     gauge_param.anisotropy = lattice.anisotropy
-    gauge_param.tadpole_coeff = tadpole_coeff
-    gauge_param.scale = -(1 + naik_epsilon) / (24 * tadpole_coeff * tadpole_coeff)
+    gauge_param.tadpole_coeff = 1.0  # tadpole_coeff
+    gauge_param.scale = 1.0  # -(1 + naik_epsilon) / (24 * tadpole_coeff * tadpole_coeff)
     gauge_param.type = QudaLinkType.QUDA_WILSON_LINKS
     gauge_param.gauge_order = QudaGaugeFieldOrder.QUDA_QDP_GAUGE_ORDER
-    gauge_param.t_boundary = lattice.t_boundary
+    gauge_param.t_boundary = QudaTboundary(lattice.t_boundary)
 
     gauge_param.gauge_fix = QudaGaugeFixed.QUDA_GAUGE_FIXED_NO
     gauge_param.ga_pad = lattice.ga_pad
@@ -238,17 +263,7 @@ def newQudaGaugeParam(
     return gauge_param
 
 
-def newQudaMultigridParam(
-    mass: float,
-    kappa: float,
-    geo_block_size: List[List[int]],
-    coarse_tol: float,
-    coarse_maxiter: int,
-    setup_tol: float,
-    setup_maxiter: int,
-    nu_pre: int,
-    nu_post: int,
-):
+def newQudaMultigridParam(mass: float, kappa: float, geo_block_size: List[List[int]], staggered: bool):
     mg_param = QudaMultigridParam()
     mg_inv_param = QudaInvertParam()
 
@@ -273,7 +288,7 @@ def newQudaMultigridParam(
     mg_inv_param.gcrNkrylov = 8
     mg_inv_param.use_init_guess = QudaUseInitGuess.QUDA_USE_INIT_GUESS_NO
 
-    location: QudaFieldLocation = _fieldLocation()
+    location = fieldLocation()
     mg_inv_param.input_location = location
     mg_inv_param.output_location = location
     mg_inv_param.dirac_order = QudaDiracFieldOrder.QUDA_DIRAC_ORDER
@@ -288,61 +303,78 @@ def newQudaMultigridParam(
 
     mg_param.invert_param = mg_inv_param
 
+    geo_block_size = geo_block_size + [[4, 4, 4, 4]]
+    if staggered:
+        geo_block_size = [[1, 1, 1, 1]] + geo_block_size
     n_level = len(geo_block_size)
     for i in range(n_level):
         geo_block_size[i] = geo_block_size[i] + [1] * (QUDA_MAX_DIM - len(geo_block_size[i]))
     mg_param.n_level = n_level
     mg_param.geo_block_size = geo_block_size
-    mg_param.spin_block_size = [2] + [1] * (QUDA_MAX_MG_LEVEL - 1)
-    mg_param.n_vec = [24] * QUDA_MAX_MG_LEVEL
-    mg_param.n_block_ortho = [1] * QUDA_MAX_MG_LEVEL
+    if staggered:
+        mg_param.spin_block_size = [0] + [0] + [1] * (QUDA_MAX_MG_LEVEL - 2)
+        mg_param.n_vec = [3] + [64] * (QUDA_MAX_MG_LEVEL - 1)
+        mg_param.n_block_ortho = [1] + [2] * (QUDA_MAX_MG_LEVEL - 1)
+    else:
+        mg_param.spin_block_size = [2] + [1] * (QUDA_MAX_MG_LEVEL - 1)
+        mg_param.n_vec = [24] * QUDA_MAX_MG_LEVEL
+        mg_param.n_block_ortho = [1] * QUDA_MAX_MG_LEVEL
+
+    mg_param.verbosity = [QudaVerbosity.QUDA_SILENT] * QUDA_MAX_MG_LEVEL
+    use_mma = QudaBoolean(mmaAvailable())
+    mg_param.setup_use_mma = [use_mma] * QUDA_MAX_MG_LEVEL
+    mg_param.dslash_use_mma = [use_mma] * QUDA_MAX_MG_LEVEL
 
     mg_param.setup_inv_type = [QudaInverterType.QUDA_CGNR_INVERTER] * QUDA_MAX_MG_LEVEL
     mg_param.n_vec_batch = [1] * QUDA_MAX_MG_LEVEL
     mg_param.num_setup_iter = [1] * QUDA_MAX_MG_LEVEL
-    mg_param.setup_tol = [setup_tol] * QUDA_MAX_MG_LEVEL
-    mg_param.setup_maxiter = [setup_maxiter] * QUDA_MAX_MG_LEVEL
-    mg_param.setup_maxiter_refresh = [setup_maxiter // 10] * QUDA_MAX_MG_LEVEL
+    mg_param.setup_tol = [1e-6] * QUDA_MAX_MG_LEVEL
+    mg_param.setup_maxiter = [1000] * QUDA_MAX_MG_LEVEL
+    mg_param.setup_maxiter_refresh = [1000 // 10] * QUDA_MAX_MG_LEVEL
     mg_param.setup_type = QudaSetupType.QUDA_NULL_VECTOR_SETUP
     mg_param.pre_orthonormalize = QudaBoolean.QUDA_BOOLEAN_FALSE
     mg_param.post_orthonormalize = QudaBoolean.QUDA_BOOLEAN_TRUE
 
     mg_param.coarse_solver = [QudaInverterType.QUDA_GCR_INVERTER] * (n_level - 1) + [
         QudaInverterType.QUDA_CA_GCR_INVERTER
-    ] * (QUDA_MAX_MG_LEVEL - n_level + 1)
-    mg_param.coarse_solver_tol = [coarse_tol] * QUDA_MAX_MG_LEVEL
-    mg_param.coarse_solver_maxiter = [coarse_maxiter] * QUDA_MAX_MG_LEVEL
-    mg_param.transfer_type = [QudaTransferType.QUDA_TRANSFER_AGGREGATE] * QUDA_MAX_MG_LEVEL
+    ] * (QUDA_MAX_MG_LEVEL - (n_level - 1))
+    mg_param.coarse_solver_tol = [0.25] * QUDA_MAX_MG_LEVEL
+    mg_param.coarse_solver_maxiter = [16] * QUDA_MAX_MG_LEVEL
+    # mg_param.coarse_solver_ca_basis_size = [16] * QUDA_MAX_MG_LEVEL
 
     mg_param.smoother = [QudaInverterType.QUDA_CA_GCR_INVERTER] * QUDA_MAX_MG_LEVEL
     mg_param.smoother_tol = [0.25] * QUDA_MAX_MG_LEVEL
-    mg_param.nu_pre = [nu_pre] * QUDA_MAX_MG_LEVEL
-    mg_param.nu_post = [nu_post] * QUDA_MAX_MG_LEVEL
+    # mg_param.smoother_tol = [1e-10] * QUDA_MAX_MG_LEVEL
+    mg_param.nu_pre = [0] * QUDA_MAX_MG_LEVEL
+    mg_param.nu_post = [8] * QUDA_MAX_MG_LEVEL
     mg_param.omega = [1.0] * QUDA_MAX_MG_LEVEL
     mg_param.smoother_schwarz_type = [QudaSchwarzType.QUDA_INVALID_SCHWARZ] * QUDA_MAX_MG_LEVEL
     mg_param.smoother_schwarz_cycle = [1] * QUDA_MAX_MG_LEVEL
 
-    mg_param.coarse_grid_solution_type = [QudaSolutionType.QUDA_MATPC_SOLUTION] * QUDA_MAX_MG_LEVEL
-    mg_param.smoother_solve_type = [QudaSolveType.QUDA_DIRECT_PC_SOLVE] * QUDA_MAX_MG_LEVEL
+    coarse_grid_solution_type = [QudaSolutionType.QUDA_MATPC_SOLUTION] * QUDA_MAX_MG_LEVEL
+    smoother_solve_type = [QudaSolveType.QUDA_DIRECT_PC_SOLVE] * QUDA_MAX_MG_LEVEL
+    if staggered:
+        coarse_grid_solution_type[1] = QudaSolutionType.QUDA_MAT_SOLUTION
+        smoother_solve_type[1] = QudaSolveType.QUDA_DIRECT_SOLVE
+    mg_param.coarse_grid_solution_type = coarse_grid_solution_type
+    mg_param.smoother_solve_type = smoother_solve_type
     mg_param.cycle_type = [QudaMultigridCycleType.QUDA_MG_CYCLE_RECURSIVE] * QUDA_MAX_MG_LEVEL
     mg_param.global_reduction = [QudaBoolean.QUDA_BOOLEAN_TRUE] * QUDA_MAX_MG_LEVEL
 
-    mg_param.mu_factor = [1.0] * QUDA_MAX_MG_LEVEL
-
     mg_param.location = [QudaFieldLocation.QUDA_CUDA_FIELD_LOCATION] * QUDA_MAX_MG_LEVEL
     mg_param.setup_location = [QudaFieldLocation.QUDA_CUDA_FIELD_LOCATION] * QUDA_MAX_MG_LEVEL
-    mg_param.setup_minimize_memory = QudaBoolean.QUDA_BOOLEAN_FALSE
+    # mg_param.setup_minimize_memory = QudaBoolean.QUDA_BOOLEAN_FALSE
     mg_param.compute_null_vector = QudaComputeNullVector.QUDA_COMPUTE_NULL_VECTOR_YES
     mg_param.generate_all_levels = QudaBoolean.QUDA_BOOLEAN_TRUE
     mg_param.run_verify = QudaBoolean.QUDA_BOOLEAN_TRUE
     mg_param.run_low_mode_check = QudaBoolean.QUDA_BOOLEAN_FALSE
     mg_param.run_oblique_proj_check = QudaBoolean.QUDA_BOOLEAN_FALSE
 
-    use_mma: QudaBoolean = _useMMA()
-    mg_param.setup_use_mma = [use_mma] * QUDA_MAX_MG_LEVEL
-    mg_param.dslash_use_mma = [use_mma] * QUDA_MAX_MG_LEVEL
-
-    mg_param.verbosity = [QudaVerbosity.QUDA_SILENT] * QUDA_MAX_MG_LEVEL
+    mg_param.mu_factor = [1.0] * QUDA_MAX_MG_LEVEL
+    transfer_type = [QudaTransferType.QUDA_TRANSFER_AGGREGATE] * QUDA_MAX_MG_LEVEL
+    if staggered:
+        transfer_type[0] = QudaTransferType.QUDA_TRANSFER_OPTIMIZED_KD
+    mg_param.transfer_type = transfer_type
 
     return mg_param, mg_inv_param
 
@@ -354,7 +386,7 @@ def newQudaInvertParam(
     maxiter: int,
     clover_coeff: float,
     clover_anisotropy: float,
-    mg_param: QudaMultigridParam,
+    mg_param: Optional[QudaMultigridParam],
 ):
     invert_param = QudaInvertParam()
 
@@ -389,7 +421,7 @@ def newQudaInvertParam(
     invert_param.gcrNkrylov = 8
     invert_param.use_init_guess = QudaUseInitGuess.QUDA_USE_INIT_GUESS_NO
 
-    location: QudaFieldLocation = _fieldLocation()
+    location = fieldLocation()
     invert_param.input_location = location
     invert_param.output_location = location
     invert_param.dirac_order = QudaDiracFieldOrder.QUDA_DIRAC_ORDER
@@ -421,13 +453,14 @@ def newQudaInvertParam(
 
 
 def loadClover(
-    clover: LatticeClover,
-    clover_inv: LatticeClover,
-    gauge: LatticeGauge,
+    clover: Optional[LatticeClover],
+    clover_inv: Optional[LatticeClover],
+    gauge: Optional[LatticeGauge],
     gauge_param: QudaGaugeParam,
     invert_param: QudaInvertParam,
 ):
     if clover is None or clover_inv is None:
+        assert gauge is not None
         clover_anisotropy = invert_param.clover_csw
         t_boundary = gauge_param.t_boundary
         anisotropy = gauge_param.anisotropy
@@ -498,7 +531,26 @@ def loadGauge(gauge: LatticeGauge, gauge_param: QudaGaugeParam):
     gauge_param.use_resident_gauge = 1
 
 
-def newPathCoeff(tadpole_coeff: float):
+def newAsqtadPathCoeff(tadpole_coeff: float):
+    u1 = 1.0 / tadpole_coeff
+    u2 = u1 * u1
+    u4 = u2 * u2
+    u6 = u4 * u2
+
+    path_coeff_1 = [
+        ((1.0 / 8.0) + (6.0 / 16.0) + (1.0 / 8.0)),  # one link
+        # One link is 1/8 as in fat7 +3/8 for Lepage + 1/8 for Naik
+        u2 * (-1.0 / 24.0),  # Naik
+        u2 * (-1.0 / 8.0) * 0.5,  # simple staple
+        u4 * (1.0 / 8.0) * 0.25 * 0.5,  # displace link in two directions
+        u6 * (-1.0 / 8.0) * 0.125 * (1.0 / 6.0),  # displace link in three directions
+        u4 * (-1.0 / 16.0),  # Correct O(a^2) errors
+    ]
+
+    return numpy.array(path_coeff_1, "<f8")
+
+
+def newHISQPathCoeff(tadpole_coeff: float):
     u1 = 1.0 / tadpole_coeff
     u2 = u1 * u1
     u4 = u2 * u2
@@ -554,14 +606,25 @@ def computeULink(gauge: LatticeGauge, gauge_param: QudaGaugeParam):
     return u_link
 
 
-def computeWLink(
-    u_link: LatticeGauge, return_v_link: bool, path_coeff: NDArray[numpy.float64], gauge_param: QudaGaugeParam
-):
-    v_link = LatticeGauge(u_link.latt_info) if return_v_link else None
+def computeWLink(u_link: LatticeGauge, path_coeff: NDArray[numpy.float64], gauge_param: QudaGaugeParam):
     w_link = LatticeGauge(u_link.latt_info)
-
     computeKSLinkQuda(
-        v_link.data_ptrs if return_v_link else nullptrs,
+        nullptrs,
+        nullptrs,
+        w_link.data_ptrs,
+        u_link.data_ptrs,
+        path_coeff,
+        gauge_param,
+    )
+
+    return w_link
+
+
+def computeVWLink(u_link: LatticeGauge, path_coeff: NDArray[numpy.float64], gauge_param: QudaGaugeParam):
+    v_link = LatticeGauge(u_link.latt_info)
+    w_link = LatticeGauge(u_link.latt_info)
+    computeKSLinkQuda(
+        v_link.data_ptrs,
         nullptrs,
         w_link.data_ptrs,
         u_link.data_ptrs,
@@ -579,7 +642,6 @@ def computeXLink(
 ):
     fatlink = LatticeGauge(w_link.latt_info)
     longlink = LatticeGauge(w_link.latt_info)
-
     computeKSLinkQuda(
         fatlink.data_ptrs,
         longlink.data_ptrs,
@@ -600,10 +662,9 @@ def computeXLinkEpsilon(
     naik_epsilon: float,
     gauge_param: QudaGaugeParam,
 ):
-    fatlink_epsilon = LatticeGauge(w_link.latt_info) if naik_epsilon != 0 else None
-    longlink_epsilon = LatticeGauge(w_link.latt_info) if naik_epsilon != 0 else None
-
     if naik_epsilon != 0:
+        fatlink_epsilon = LatticeGauge(w_link.latt_info)
+        longlink_epsilon = LatticeGauge(w_link.latt_info)
         computeKSLinkQuda(
             fatlink_epsilon.data_ptrs,
             longlink_epsilon.data_ptrs,
@@ -616,6 +677,7 @@ def computeXLinkEpsilon(
         longlink_epsilon *= naik_epsilon
         fatlink_epsilon += fatlink
         longlink_epsilon += longlink
+
         return fatlink_epsilon, longlink_epsilon
     else:
         return fatlink, longlink
@@ -631,26 +693,33 @@ def loadStaggeredGauge(gauge: LatticeGauge, gauge_param: QudaGaugeParam):
 def loadFatLongGauge(
     fatlink: LatticeGauge,
     longlink: LatticeGauge,
+    tadpole_coeff: float,
+    naik_epsilon: float,
     gauge_param: QudaGaugeParam,
 ):
+    ga_pad = gauge_param.ga_pad
     staggered_phase_type = gauge_param.staggered_phase_type
 
     gauge_param.use_resident_gauge = 0
     gauge_param.type = QudaLinkType.QUDA_ASQTAD_FAT_LINKS
     loadGaugeQuda(fatlink.data_ptrs, gauge_param)
     gauge_param.type = QudaLinkType.QUDA_ASQTAD_LONG_LINKS
-    gauge_param.ga_pad = gauge_param.ga_pad * 3
+    setReconstructParam(_reconstruct["staggered"], gauge_param)
+    gauge_param.ga_pad = ga_pad * 3
     gauge_param.staggered_phase_type = QudaStaggeredPhase.QUDA_STAGGERED_PHASE_NO
+    gauge_param.tadpole_coeff = tadpole_coeff
+    gauge_param.scale = -(1 + naik_epsilon) / (24 * tadpole_coeff * tadpole_coeff)
     loadGaugeQuda(longlink.data_ptrs, gauge_param)
     gauge_param.type = QudaLinkType.QUDA_WILSON_LINKS
-    gauge_param.ga_pad = gauge_param.ga_pad // 3
+    setReconstructParam(_reconstruct["none"], gauge_param)
+    gauge_param.ga_pad = ga_pad
     gauge_param.staggered_phase_type = staggered_phase_type
+    gauge_param.tadpole_coeff = 1.0
+    gauge_param.scale = 1.0
     gauge_param.use_resident_gauge = 1
 
 
 def performance(invert_param: QudaInvertParam):
-    from .. import getLogger
-
     gflops, secs = invert_param.gflops, invert_param.secs
     getLogger().info(f"Time = {secs:.3f} secs, Performance = {gflops / secs:.3f} GFLOPS")
 
